__author__ = "Tobias O, Michael H., Celine M., Duc Anh L."
__copyright__ = "Copyright 2022, Rostlab"
__license__ = "AGPL-3.0-or-later"
__version__ = "1.0.0"
__maintainer__ = ["Duc Anh L.", "Tobias O."]
__email__ = ""
__status__ = "Production"
